from koslab.messengerbot.request import Response
from koslab.messengerbot.logger import logger
from kombu import Connection, Exchange, Queue
from multiprocessing import Process, Pool
import json
import signal

__all__ = ['Bots', 'KombuBots']

def spawn_bot(bot_class, bot_args, event):
    bot = bot_class(**bot_args)
    bot.handle_event(event)

def spawn_bot_amqp(bot_class, bot_args, send_transport, send_exchange, send_queue, 
                    event, message):
    bot = bot_class(send_transport=send_transport, send_exchange=send_exchange, 
            send_queue=send_queue, **bot_args)
    bot.handle_event(event)
    message.ack()

def spawn_send_message_worker(event, message):
    from koslab.messengerbot.bot import send_message
    resp = send_message(**event)
    if resp.status_code == 200:
        message.ack()

class Bots(object):
    '''
    :param: validation_token: hub validation token
    :param: page_bots: a dictionary of page id and MessengerBot implementation
    '''
    def __init__(self, validation_token, page_bots):
        self.validation_token = validation_token
        self.page_bots = page_bots

    def webhook(self, request):
        if request.method == 'GET':
            return self.webhook_get(request)
        if request.method == 'POST':
            return self.webhook_post(request)

    def webhook_get(self, request):
        if (request.get('hub.mode') == 'subscribe' and 
            request.get('hub.verify_token') == self.validation_token):
            logger.info('Received hub challenge')
            return Response(body=request.get('hub.challenge'), status=200)
        else:
            logger.info('Invalid hub challenge')
            return Response(status=403)

    def webhook_post(self, request):
        data = json.loads(request.body)

        if (data['object'] == 'page'):
            for entry in data['entry']:
                page_id = entry['id']
                timestamp = entry['time']
                bot_class, bot_args = self.get_bot(page_id)
                for event in entry['messaging']:
                    p = Process(target=spawn_bot, 
                            args=(bot_class, bot_args, event))
                    p.start()
                    def stop_bot(signum, frame):
                        p.terminate()
                        p.join()
                    signal.signal(signal.SIGTERM, stop_bot)

        return Response(status=200)

    def get_bot(self, page_id):
        bot = self.page_bots.get(page_id, None)
        if bot is None:
            raise ValueError('Unable to select bot for page %s ' % page_id)
        return bot

    def init_bots(self):
        for page_id, (bot_class, bot_args) in self.page_bots.items():
            print "Configuring %s for PageID:%s" % (
                    bot_class.__name__, page_id)
            bot = bot_class(**bot_args)
            bot.configure()

    def initialize(self):
        self.init_bots()

class KombuBots(Bots):
    '''
    Web hook for Kombu queue

    :param: validation_token: hub validation token
    '''
    def __init__(self, validation_token, page_bots, transport=None, 
                exchange=None, queue=None, send_transport=None,
                send_exchange=None, send_queue=None):
        super(KombuBots, self).__init__(validation_token, page_bots)
        self.page_bots = page_bots
        self.transport = transport or 'amqp://guest:guest@localhost:5672//'
        self.exchange = exchange or 'MessengerBot'
        self.queue = queue or 'messages'
        self.send_transport = send_transport or transport
        self.send_exchange = send_exchange or 'MessengerBot'
        self.send_queue = send_queue or 'replies'

    def webhook_post(self, request):
        data = json.loads(request.body)
        if (data['object'] == 'page'):
            for entry in data['entry']:
                page_id = entry['id']
                timestamp = entry['time']
                self.queue_events(entry['messaging'])
        return Response(status=200)

    def queue_events(self, events):
        exchange = Exchange(self.exchange, 'direct', durable=True)
        queue = Queue(self.queue, exchange=exchange, routing_key=self.queue)
        with Connection(self.transport) as conn:
            producer = conn.Producer(serializer='json')
            for event in events:
                producer.publish(event, exchange=exchange, 
                        routing_key=queue.routing_key,
                        declare=[queue])

    def consume(self):
        def worker(event, message):
            page_id = event['recipient']['id']
            bot_class, bot_args = self.get_bot(page_id)
            p = Process(target=spawn_bot_amqp, args=(bot_class, bot_args, 
                        self.transport, self.send_exchange, 
                        self.send_queue, event, message))
            p.start()
            def stop_worker(signum, frame):
                p.terminate()
                p.join()
            signal.signal(signal.SIGTERM, stop_worker)

        exchange = Exchange(self.exchange, 'direct', durable=True)
        queue = Queue(self.queue, exchange=exchange, routing_key=self.queue)
        with Connection(self.transport) as conn:
            with conn.Consumer(queue, callbacks=[worker]) as consumer:
                while True:
                    conn.drain_events()

    def sender(self):
        def worker(event, message):
            p = Process(target=spawn_send_message_worker, args=(event, message))
            p.start()
            def stop_worker(signum, frame):
                p.terminate()
                p.join()
            signal.signal(signal.SIGTERM, stop_worker)


        exchange = Exchange(self.send_exchange, 'direct', durable=True)
        queue = Queue(self.send_queue, exchange=exchange, 
                        routing_key=self.send_queue)
        with Connection(self.send_transport) as conn:
            with conn.Consumer(queue, callbacks=[worker]) as consumer:
                while True:
                    conn.drain_events()

    def initialize(self):
        super(KombuBots, self).init_bots()
        print "Running %s message consumer" % self
        p1 = Process(target=self.consume)
        p1.start()
        print "Running %s message sender" % self
        p2 = Process(target=self.sender)
        p2.start()
        def stop_sender_consumer(signum, frame):
            p1.terminate()
            p1.join()
            p2.terminate()
            p2.join()
        signal.signal(signal.SIGTERM, stop_sender_consumer)



import morepath
from koslab.messengerbot.hub import App
import yaml
import sys
import os

def start_hub():
    if len(sys.argv) == 1:
        print "%s [config.yml]" % sys.argv[0]
        sys.exit(1)

    sys.path.append(os.getcwd())
    config = yaml.load(open(sys.argv[1]).read())
    @App.setting_section(section='config')
    def config_section():
        return config

    morepath.run(App(config), ignore_cli=True)



from beaker.cache import CacheManager

class SessionManager(object):

    def __init__(self, type='memory', **kwargs):
        opts = kwargs or {
            'data_dir': '/tmp/messengerbot-cache/data',
            'lock_dir': '/tmp/messengerbot-cache/lock'
        }
        opts['type'] = type
        self.cachemgr = CacheManager(**opts)

    def get_session(self, event):
        ns = '.'.join([event['recipient']['id'],event['sender']['id']])
        cache = self.cachemgr.get_cache(ns)
        return Session(cache)

_marker = []

class Session(object):

    def __init__(self, cache):
        self.cache = cache

    def get(self, key, default=_marker):
        if default is not _marker:
            value = self.cache._get_value(key)
            if not value.has_current_value():
                return default
        return self.cache.get(key=key)

    def set(self, key, value):
        self.cache.set_value(key, value)

    def delete(self, key):
        self.cache.remove_value(key=key)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __delitem__(self, key):
        return self.delete(key)



import requests
from koslab.messengerbot.logger import logger
import json
import re
import time
from koslab.messengerbot.session import SessionManager
from beaker.util import parse_cache_config_options
from kombu import Connection, Exchange, Queue
import os

def send_message(page_access_token, recipient, message=None, sender_action=None):
    request_data = { 'recipient': recipient }
    if not (message or sender_action):
        raise ValueError('message or sender_action is required')
    if message:
        request_data['message'] = message
    elif sender_action:
        if sender_action not in ['mark_seen', 'typing_on', 'typing_off']:
            raise ValueError('Invalid Action %s' % sender_action)
        request_data['sender_action'] = sender_action

    url = 'https://graph.facebook.com/v2.6/me/messages'
    resp = requests.post('%s?access_token=%s' % (url,
                page_access_token), json=request_data)
    resp_data = resp.json()
    error = resp_data.get('error', None)
    if error and error['code'] != 2:
        # rate limit error, retry
        if error['code'] == 4:
            time.sleep(5)
            self.send(recipient, message, sender_action)
        else:
            logger.error('Error sending message: %s' % json.dumps(error))
    return resp


class BaseMessengerBot(object):

    POSTBACK_HANDLERS = {}
    GREETING_TEXT = 'Hello World!'
    STARTUP_MESSAGE = { 'text': 'Hello World!' }
    PERSISTENT_MENU = [{
       'type': 'postback',
       'title': 'Get Started',
       'payload': 'messengerbot.get_started'
    }]


    def __init__(self, page_access_token, session_opts=None, 
                 send_transport=None, send_exchange=None, send_queue=None):
        self.session_opts = session_opts or {
            'type': 'file',
            'data_dir': '/tmp/messengerbot-cache/data',
            'lock_dir': '/tmp/messengerbot-cache/lock'
        }
        self.sessionmgr = SessionManager(**self.session_opts)
        self.send_transport = send_transport
        self.page_access_token = page_access_token
        self.send_exchange = send_exchange or 'MessengerBot'
        self.send_queue = send_queue or 'replies'


    def start_hook(self, event):
        if self.STARTUP_MESSAGE is not None:
            self.reply(event, self.STARTUP_MESSAGE)

    def authentication_hook(self, event):
        pass

    def message_hook(self, event):
        pass

    def message_delivered_hook(self, event):
        pass

    def postback_hook(self, event):
        try:
            payload = json.loads(event['postback']['payload'])
            ev = payload['event']
        except:
            ev = payload = event['postback']['payload']
        if ev == 'messengerbot.get_started':
            self.start_hook(event)
        method = self.POSTBACK_HANDLERS.get(ev, None)
        if method:
            getattr(self, method)(event)

    def read_hook(self, event):
        pass

    def account_linking_hook(self, event):
        pass

    def user_profile(self, user, fields=None):
        fields = fields or ['first_name','last_name','profile_pic','locale',
                                'timezone','gender']
        url = 'https://graph.facebook.com/v2.6/%s' % user['id']
        resp = requests.get(url, data={'fields': ','.join(fields)})
        resp_data = resp.json()
        if resp_data.get('error', None):
            return {
                'first_name': '',
                'last_name': '',
                'profile_pic': '',
                'locale': 'en_US',
                'timezone': 0,
                'gender': ''
            }
        return resp_data

    def send(self, recipient, message=None, sender_action=None):
        if self.send_transport is not None:
            return self.queue_send(
                            recipient=recipient, message=message,
                            sender_action=sender_action)
        send_message(self.page_access_token, recipient=recipient, 
                message=message, sender_action=sender_action)

    def queue_send(self, recipient, message=None, sender_action=None):
        exchange = Exchange(self.send_exchange, 'direct', durable=True)
        queue = Queue(self.send_queue, exchange=exchange, 
                        routing_key=self.send_queue)
        with Connection(self.send_transport) as conn:
            producer = conn.Producer(serializer='json')
            event = {
                'recipient': recipient,
                'message': message,
                'sender_action': sender_action,
                'page_access_token': self.page_access_token
            }
            producer.publish(event, exchange=exchange, 
                                routing_key=queue.routing_key,
                                declare=[queue])

    def reply(self, event, message=None):
        self.send(recipient=event['sender'], sender_action='mark_seen')
        self.send(recipient=event['sender'], sender_action='typing_on')
        if callable(message):
            message = message(event)
        if isinstance(message, str):
            message = {'text': message}
        res = self.send(recipient=event['sender'], message=message)
        self.send(recipient=event['sender'], sender_action='typing_off')

    def handle_event(self, event):
        if event.get('optin', None):
            logger.debug('Authentication hook: %s' % json.dumps(event))
            self.authentication_hook(event)
        elif event.get('message', None):
            logger.debug('Message hook: %s' % json.dumps(event))
            self.message_hook(event)
        elif event.get('delivery', None):
            logger.debug('Message delivered hook: %s' % json.dumps(event))
            self.message_delivered_hook(event)
        elif event.get('postback', None):
            logger.debug('Postback hook: %s' % json.dumps(event))
            self.postback_hook(event)
        elif event.get('read', None):
            logger.debug('Read hook: %s' % json.dumps(event))
            self.read_hook(event)
        elif event.get('account_linking', None):
            logger.debug('Account linking hook: %s' % json.dumps(event))
            self.account_linking_hook(event)
        else:
            logger.info(
                'Webhook received unknown messagingEvent %s' % json.dumps(event))

    def thread_settings(self, setting):
        url = 'https://graph.facebook.com/v2.6/me/thread_settings'
        resp = requests.post('%s?access_token=%s' % (url,
                            self.page_access_token), json=setting)
        return resp

    def configure(self):
        # configure greeting
        if self.GREETING_TEXT is not None:
            self.thread_settings({
                'setting_type': 'greeting', 
                'greeting': {
                    'text': self.GREETING_TEXT
                }
            })
        # configure Get Started button
        self.thread_settings({
            'setting_type': 'call_to_actions',
            'thread_state': 'new_thread',
            'call_to_actions': [
                { 'payload': 'messengerbot.get_started' }
            ]
        })

        # configure Persistent Menu
        if self.PERSISTENT_MENU is not None:
            self.thread_settings({
                'setting_type': 'call_to_actions',
                'thread_state': 'existing_thread',
                'call_to_actions': self.PERSISTENT_MENU
            })

    def get_session(self, event):
        return self.sessionmgr.get_session(event)




class WebObRequestAdapter(object):
    def __init__(self, request):
        self.request = request

    def get(self, key):
        return self.request.GET.get(key)

    @property
    def body(self):
        return self.request.body

    @property
    def method(self):
        return self.request.method

class Response(object):
    def __init__(self, body=None, status=None, headerlist=None,
            content_type=None):
        self.body = body
        self.status = status
        self.headerlist = headerlist
        self.content_type = content_type

    def params(self):
        return {
            'body': self.body,
            'status': self.status,
            'headerlist': self.headerlist,
            'content_type': self.content_type
        }



from setuptools import setup, find_packages
import os

version = '1.0b4'

long_description = (
    open('README.rst').read()
    + '\n' +
    'Contributors\n'
    '============\n'
    + '\n' +
    open('CONTRIBUTORS.txt').read()
    + '\n' +
    open('CHANGES.txt').read()
    + '\n')

setup(name='koslab.messengerbot',
      version=version,
      description="Facebook messenger bot framework",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        ],
      keywords='morepath facebook bot messenger chatbot',
      author='Izhar Firdaus',
      author_email='kagesenshi.87@gmail.com',
      url='http://github.com/koslab/koslab.messengerbot/',
      license='MIT',
      packages=find_packages('src'),
      package_dir = {'': 'src'},
      namespace_packages=['koslab'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'kombu',
          'requests',
          'beaker',
          'morepath',
          'PyYAML'
          # -*- Extra requirements: -*-
      ],
      entry_points={
          'console_scripts':
              'messengerbot_hub=koslab.messengerbot.scripts:start_hub'
      })



import morepath
import logging
from koslab.messengerbot.bots import Bots, KombuBots
from koslab.messengerbot.request import WebObRequestAdapter
import importlib

class App(morepath.App):

    def __init__(self, config):
        params = {
            'validation_token': config['hub_verify_token'],
            'page_bots': self.bots_config(config)
        }
        if config.get('use_message_queue', False):
            coordinator_klass = KombuBots
            params['transport'] = config.get('message_queue', None)
        else:
            coordinator_klass = Bots

        self.bots = coordinator_klass(**params)
        print("=== Configuring bots ===")
        self.bots.initialize()
        print("=== Starting up webhook on /%s ===" % config['webhook'])

    def bots_config(self, config):
        bots = config.get('bots', [])
        result = {}
        for bot in bots:
            bot_module, bot_class = bot.get('class').strip().split(':')
            m = importlib.import_module(bot_module)
            klass = getattr(m, bot_class)
            kwargs = bot.get('args', {})
            kwargs['page_access_token'] = bot['access_token']
            result[str(bot['page_id'])] = (klass, kwargs)
        return result

class Root(object):
    pass

@App.path(path='{webhook}', model=Root)
def get_root(webhook):
    config = morepath.settings().config
    if config.webhook != webhook:
        return None
    return Root()

@App.view(model=Root, name='', request_method='GET')
def webhook_get(context, request):
    req = WebObRequestAdapter(request)
    resp = request.app.bots.webhook(req)
    return morepath.Response(**resp.params())

@App.view(model=Root, name='', request_method='POST')
def webhook_post(context, request):
    req = WebObRequestAdapter(request)
    resp = request.app.bots.webhook(req)
    return morepath.Response(**resp.params())

